from dune.perftool.generation import instruction
from dune.perftool.loopy.target import dtype_floatingpoint
from dune.perftool.options import get_option
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.argument import name_accumulation_variable
from dune.perftool.pdelab.localoperator import boundary_predicates
from dune.perftool.generation.loopy import function_mangler, globalarg
import loopy as lp
import pymbolic.primitives as prim


def name_accumulation_alias(container, accumspace):
    name = container+"_"+accumspace.lfs.name+"_alias"
    name_tail = container+"_"+accumspace.lfs.name+"_alias_tail"
    k = get_option("number_of_blocks")
    p = accumspace.element.degree()

    def _add_alias_insn(name):
        dim = world_dimension()
        element_stride = tuple(p*(p*k+1)**i for i in range(0,dim))
        index_stride = tuple((p*k+1)**i for i in range(0,dim))
        globalarg(name, shape=(k,)*dim+(p + 1,)*dim, strides=element_stride+index_stride, managed=True)
        code = "auto {} = &{}.container()({},0);".format(name, container, accumspace.lfs.name)
        instruction(within_inames=frozenset(),
                    code=code,
                    read_variables=frozenset({container}),
                    assignees=frozenset({name}))

    _add_alias_insn(name)
    _add_alias_insn(name_tail)
    return name


@function_mangler
def residual_weight_mangler(knl, func, arg_dtypes):
    if isinstance(func, str) and func.endswith('.weight'):
        return lp.CallMangleInfo(func, (lp.types.NumpyType(dtype_floatingpoint()),), ())


def generate_accumulation_instruction(expr, visitor):
    # Collect the lfs and lfs indices for the accumulate call
    test_lfs = determine_accumulation_space(visitor.test_info, 0)
    # In the jacobian case, also determine the space for the ansatz space
    ansatz_lfs = determine_accumulation_space(visitor.trial_info, 1)

    # Collect the lfs and lfs indices for the accumulate call
    accumvar = name_accumulation_variable(test_lfs.get_restriction() + ansatz_lfs.get_restriction())
    accumvar_alias = name_accumulation_alias(accumvar, test_lfs)

    predicates = boundary_predicates(expr, visitor.measure, visitor.subdomain_id)

    quad_inames = visitor.interface.quadrature_inames()
    lfs_inames = visitor.test_info.inames
    if visitor.trial_info:
        lfs_inames = lfs_inames + visitor.trial_info.inames

    assignee = prim.Subscript(prim.Variable(accumvar_alias), tuple(prim.Variable(i) for i in lfs_inames))

    expr_with_weight = prim.Product((expr, prim.Call(prim.Variable(accumvar+'.weight'),())))

    instruction(assignee=assignee,
                expression=prim.Sum((expr_with_weight, assignee)),
                forced_iname_deps=frozenset(lfs_inames).union(frozenset(quad_inames)),
                forced_iname_deps_is_final=True,
                predicates=predicates,
                tags=frozenset({'accum'})
                )
