""" Iname duplication strategies to make kernels schedulable """
from dune.perftool.options import get_option
from loopy import (has_schedulable_iname_nesting,
                   get_iname_duplication_options,
                   duplicate_inames,
                   )


def heuristic_duplication(kernel):
    # If the given kernel is schedulable, nothing needs to be done.
    if has_schedulable_iname_nesting(kernel):
        return kernel

    # List all duplication options and return the transformed
    # kernel if one such duplication transformation was enough to solve the problem.
    for iname, within in get_iname_duplication_options(kernel):
        dup_kernel = duplicate_inames(kernel, iname, within)
        if has_schedulable_iname_nesting(dup_kernel):
            return dup_kernel

    if get_option('blockstructured'):
        duplication_options = dict(get_iname_duplication_options(kernel))
        suffixes = ['x','y','z']
        for iname in duplication_options:
            if iname[-1] in suffixes:
                dup_kernel = kernel
                base = iname[:-1]
                for s in suffixes:
                    if base+s in duplication_options:
                        dup_kernel = duplicate_inames(dup_kernel, base+s, duplication_options[base+s])
                        del duplication_options[base+s]
                if has_schedulable_iname_nesting(dup_kernel):
                    return dup_kernel

    raise NotImplementedError("Your kernel needs multiple iname duplications! No generic algorithm implemented for that yet! (#39)")
